#!/usr/bin/perl -w
# $Id: prepumbrella.pl,v 1.35 2005/10/09 22:13:42 oliver Exp $

use strict;
use Getopt::Long;
use File::Basename;
use File::Copy;
use File::Path;
use FileHandle;
#use FindBin;
#use lib $FindBin::Bin; # Messages.pm is in script directory
                       # add other lib dirs in environment var PERL5LIB
use lib "/sansom/gfio/oliver/Devel/Setup/UmbrellaSampling";
use Messages qw($DEBUG open_log errmsg msg logger warning date Progname);
# Not using XML (yet) but probably I should turn the the configuration data
# file into an xml file. Would make things easier in the future...
# use XML::Simple qw(:strict) ;
use Data::Dumper;

my $program_version = q( $Id: prepumbrella.pl,v 1.35 2005/10/09 22:13:42 oliver Exp $ );
my $progname = Progname();
my $gmx_version = "3.2.1-lmb";
my ($XX,$YY,$ZZ) = (0,1,2);           # use names for dimensions in arrays

my $TOPOLOGY = "topol.top";
my $CONFSTART = "conf.dat";
my $MDP = "grompp.mdp";
my $INDEX = "pull.ndx";
my $NUM_PROCS = 1;

my $QSUB_ALL = "array_%s.sh";         # script which qsubs all windows QSCRIPT
my $QSCRIPT =  "umbrella.sh";         # name of individual qsub script; 
                                      # prefixed with the TARGETSYSTEM
my $PROJECT = ".";
my $PREFIX = "";
my $OUTFORMAT = '%s%+.4f';

my $K_UMBRELLA;           # kJ mol^-1 nm^-2
my $TYPE;                 # particle type (name from gro file)
my $TEMPERATURE = 298;    # Kelvin (same default as g_wham)
my $BARRIER = 1.5;        # kT; potential at ~centre of adjacent window
my ($N_WINDOWS) = (20);
my ($MIN,$MAX,$LENGTH);
my @TARGETSYSTEM = ();
my $TargetSystemDefault = "local";
my ($show_help,$show_targets,$show_config,$show_version,$opt_verbose);

my %requiredfiles = ("--topology"  => \$TOPOLOGY, 
                     "--confstart" => \$CONFSTART, 
                     "--mdp"       => \$MDP,
                     "--index"     => \$INDEX,
                     );
my %requiredvars = ( "--type"      => \$TYPE,
                     "--np"        => \$NUM_PROCS,
                     );

my %UmbrellaParams;   # stores global parameters set through the configuration list
my %configvars = (    # allowed vars in UmbrellaParams ('required' will be changed to
                   "center_of_mass" => {"required" => 1, # 1 once the var is read in) 
                                        "isset"    => 0,  
                                        "dim"      => 3},
                   "reference_group" => {"required" => 0,
                                         "isset"    => 0,
                                         "missing"  => 0,
                                         "default"  => "protein",
                                         "dim"      => 1},
                   "reftype"         => {"required" => 0,
                                         "isset"    => 0,
                                         "missing"  => 0,
                                         "default"  => "com_t0",
                                         "dim"      => 1},
                   "KConfine"        => {"required" => 0,
                                         "isset"    => 0,
                                         "missing"  => 0,
                                         "default"  => "4500.0",
                                         "dim"      => 1},
                   "RConfine"        => {"required" => 0,
                                         "isset"    => 0,
                                         "missing"  => 0,
                                         "default"  => "0.0",
                                         "dim"      => 1},                   
                   );

# Gromacs programs used while running prepumbrella.pl

# sort and shuffle should give a sizable (?) improvement on SMP but
# cannot be used because the index files which determine the molecule
# to be restrained were made BEFORE the shuffling takes place

my %Gromacs = ( "grompp" => "grompp",
                "GMXRC"  => "source /sansom/fedpacks/opt/gromacs/3.2.1/i686-pc-linux-gnu/bin/GMXRC",
                );


# queuing systems that we know about
#
# We try to automatically build run scripts for different environments.
# If there is no 'qsystem' entry for a particular targetsystem then 
# all stuff pertaining to queuing is skipped.
# Name-mangling for PBS because it has a limit of 16 chars.
my %qparams = ("SGE" => { "prefix" => '#$',   
			  "namemangle" => sub { return $_[0] },
			  "arrayjob" => 1,
			  "arraycode" => qq(run_d=\$(dirname \${job[\${SGE_TASK_ID}]})\n)
			      . qq(script=\$(basename \${job[\${SGE_TASK_ID}]})\n)
			      . qq(cd \${run_d} || { echo "Failed to cd \${run_d}. Abort."; exit 1; }\n)
			      . qq(chmod a+x \${script}\n)
			      . qq(source ./\${script}),
			},
	       "SGEnoarray" => { "prefix" => '#$',    # fudge synapse, argh
			  "namemangle" => sub { return $_[0] },
			  "arrayjob" => 0,
			},
	       "PBS" => { "prefix" => '#PBS', 
			  "namemangle" => sub { return substr($_[0],0,15) },
			  "arrayjob" => 0,
			},
	       "none" => { "prefix" => '',   
			   "namemangle" => sub { return $_[0] },
			   "arrayjob" => 0,
			 },
	      );      

# gromacs programs and things related to the run
#
# should be a xml file read by XML::Simple
# We define components of a run/queue script for different environments/machines:
# 1) torpedo (PBS)
# 2) workstations (SGE)
# 3) synapse (SGE/SCore)
# 4) zuse (array jobs)
# 5) neuron (SGE/SCore, array jobs)
# x) local
# * Only defined entries will trigger an entry in a run script 
#   (not quite working...)
# * SGE produces '#$ -X' options, PBS '#PBS -X' (see run_script() for the magic)
# * 'magical' templates are automatically replaced by values
#    NPROC       number of processors to run on
#    NAME        jobname (truncated to 15 chars to satisfy PBS)
#    NODES       number of nodes:
#    SGE_SLOTS       number of SGE slots
# * nproc is set to NUM_PROCS later on; in principle we could use it
#   to check min/max available resources or as a machine specific default
#
# FIXME: * should have entries for single cpu and multiple cpu jobs
#          they are just too different, for SGE at least
#        * this should be classed with inheritance
#        * prologue and epilogue should be arrays with one entry per line
#          (get rid of explicit linebreaks) not a single string
my %ScriptComponents = (
                 "torpedo" => {
                     "id"    => "torpedo",
                     "GMXRC"  => 
                         "source /usr/local/gmx3.2.1/i686-pc-linux-gnu/bin/GMXRC",
                     "mdrun" => "mdrun",
                     "mdrun_mpi" => "mdrun_mpi",
                     "mpirun" => "mpirun C -c2c -np NPROC",
                     "start_MPI" => "lamboot -v \$NODEFILE",
                     "nproc"  => 1,
                     "procpernode" => 2,
                     "qsub"   => "qsub -p 10",
                     "qsystem"  => {
                         "id"   => "PBS",
			 "PE" => "LAM",
                         "options"=> [ "-N NAME", 
                                       "-l nodes=NODES:ppn=PROCPERNODE:sse",
                                       "-S /bin/bash", "-r y", "-j oe" ],
                     },
                     "prologue" => 
                         qq(function die () {\n echo "\$1"; exit 1;\n})
                         .qq(\n)
                         .qq(function write_purge () {\n local scratchdir=\$1\n)
                         .q( echo -e '#!/bin/sh\nssh '${HOSTNAME}' rm -rvf '${scratchdir}'\n' > purge.sh)
                         .qq(\n)
                         .qq( chmod u+x purge.sh)
                         .qq(\n)
                         .q(})    
                         .qq(\n)
                         .qq(export PATH=\${PATH}:/usr/local/lam-6.5.2/bin)
                         .qq(\n)
                         .qq(export LAMHOME=/usr/local/lam-6.5.2)
                         .qq(\n)
                         .qq(scratchdir=/scratch/\${LOGNAME}/\${PBS_JOBID})
                         .qq(\n)
                         .qq(cd \${PBS_O_WORKDIR})
                         .qq(\n)                         
                         .qq(mkdir -p \${scratchdir} || die "Failed to make \${scratchdir} on \$(uname -n)")
                         .qq(\n)
                         .qq(write_purge \${scratchdir})
                         .qq(\n)
                         .qq(cp -v \${PBS_O_WORKDIR}/* \${scratchdir} || die "Error copying input files.")
                         .qq(\n)
                         .qq(cd \${scratchdir} || die "Failed to cd \${scratchdir} on \$(uname -n)")
                         .qq(\n),
                      "epilogue" =>
                         qq(\n)
                         .qq(lamhalt)
                         .qq(\n)
                         .qq(cp -v * \${PBS_O_WORKDIR}),
                     },
                 "torpedo_test" => {
                     "id"    => "torpedo_test",
                     "mdrun" => "mdrun",
                     "mdrun_mpi" => "mdrun_mpi",
                     "mpirun" => "mpirun C -c2c -np NPROC",
                     "start_MPI" => "lamboot -v \$NODEFILE",
                     "GMXRC"  => 
                         "source /home/oliver/opt/gromacs/3.2.1/i686-pc-linux-gnu/bin/GMXRC",
                     "nproc"  => 1,
                     "procpernode" => 2,
                     "qsub"   => "qsub -p 10",
                     "qsystem"  => {
                         "id"   => "PBS",
			 "PE" => "LAM",
                         "options"=> [ "-N NAME", 
                                       "-l nodes=NODES:ppn=PROCPERNODE:sse",
                                       "-S /bin/bash", "-r y", "-j oe" ],
                     },
                     "prologue" => 
                         qq(function die () {\n echo "\$1"; exit 1;\n})
                         .qq(\n)
                         .qq(function write_purge () {\n local scratchdir=\$1\n)
                         .q( echo -e '#!/bin/sh\nssh '${HOSTNAME}' rm -rvf '${scratchdir}'\n' > purge.sh)
                         .qq(\n)
                         .qq( chmod u+x purge.sh)
                         .qq(\n)
                         .q(})    
                         .qq(\n)
                         .qq(export PATH=\${PATH}:/usr/local/lam-6.5.2/bin)
                         .qq(\n)
                         .qq(export LAMHOME=/usr/local/lam-6.5.2)
                         .qq(\n)
                         .qq(scratchdir=/scratch/\${LOGNAME}/\${PBS_JOBID})
                         .qq(\n)
                         .qq(cd \${PBS_O_WORKDIR})
                         .qq(\n)                         
                         .qq(mkdir -p \${scratchdir} || die "Failed to make \${scratchdir} on \$(uname -n)")
                         .qq(\n)
                         .qq(write_purge \${scratchdir})
                         .qq(\n)
                         .qq(cp -v \${PBS_O_WORKDIR}/* \${scratchdir} || die "Error copying input files.")
                         .qq(\n)
                         .qq(cd \${scratchdir} || die "Failed to cd \${scratchdir} on \$(uname -n)")
                         .qq(\n),
                      "epilogue" =>
                         qq(\n)
                         .qq(lamhalt)
                         .qq(\n)
                         .qq(cp -v * \${PBS_O_WORKDIR}),
                 },
                 "synapse" => {
                     "id"    => "synapse",
                     "mdrun" => 
                         "/packages/gromacs-3.2.1-LMB/i686-pc-linux-gnu/bin/mdrun",
                     "mdrun_mpi" => qq(\$(which mdrun)),
                     "mpirun" => 
			 qq(/opt/score/bin/scout -wait -F \${HOME}/.score/ndfile.\${JOB_ID}.\${SGE_TASK_ID} -e /tmp/scrun.\${JOB_ID}.\${SGE_TASK_ID} -nodes=NODESxPROCPERNODE,noidle ),       
                     "GMXRC"  => 
                         qq(source /packages/gromacs-3.2.1-LMB/i686-pc-linux-gnu/bin/GMXRC),
                     "nproc"  => 2,
                     "procpernode" => 2,
                     "qsub"   => "qsub",
                     "qsystem" => {
                         "id" => "SGE",
			 "PE" => "SCore",
                         "options" => [ "-N NAME", "-S /bin/bash", 
                                        "-pe score SGE_SLOTS",
					"-V",
                                        qq(-masterq synapse.q), 
                                        "-r y", "-j y", "-cwd" ],
			 "prologue" => qq(\# hack for broken score_start.sh\n)
			     .qq(\# save nodefile for a task\n)
			     .qq(\# now repeat stupid   /usr/local/sge/bin/score_start.sh\n)
			     .qq(SERVER=synapse; score_dir=\${HOME}/.score;\n)
			     .qq(job_id=\${JOB_ID}.\${SGE_TASK_ID};\n)
			     .qq(ostype=\`/opt/score/deploy/config.nickname\`;\n)
			     .qq(score_bin=/opt/score/bin/bin.\${ostype};\n)
			     .qq(cat \${PE_HOSTFILE} | cut -f 1 -d " " | grep -v \${SERVER} > \${score_dir}/ndfile.\${job_id};\n)
			     .qq(echo "# Node file \${score_dir}/ndfile.\${job_id}:";\n)
			     .qq(cat \${score_dir}/ndfile.\${job_id};\n\n)
			     .qq(rm -f /tmp/scout.\${job_id};\n)
			     .qq(rm -f /tmp/scrun.\${job_id};\n)
			     .qq(ln -s \${score_bin}/scout.exe /tmp/scout.\${job_id};\n)
			     .qq(ln -s \${score_bin}/scrun.exe /tmp/scrun.\${job_id};\n\n\n),
			 "epilogue" => qq(\n\# our own cleanup\n)
			     .qq(rm -f \${score_dir}/ndfile.\${job_id};\n)
			     .qq(rm -f /tmp/scout.\${job_id};\n)
			     .qq(rm -f /tmp/scrun.\${job_id};\n),
                     },
                 },
                 "synapse_noarray" => {
                     "id"    => "synapse_noarray",
                     "mdrun" => 
                         "/packages/gromacs-3.2.1-LMB/i686-pc-linux-gnu/bin/mdrun",
                     "mdrun_mpi" => qq(\$(which mdrun)),
                     "mpirun" => 
			 qq(/opt/score/bin/scout -wait -F \${HOME}/.score/ndfile.\${JOB_ID} -e /tmp/scrun.\${JOB_ID} -nodes=NODESxPROCPERNODE,noidle ),       
                     "GMXRC"  => 
                         qq(source /packages/gromacs-3.2.1-LMB/i686-pc-linux-gnu/bin/GMXRC),
                     "nproc"  => 2,
                     "procpernode" => 2,
                     "qsub"   => "qsub",
                     "qsystem" => {
                         "id" => "SGEnoarray",
			 "PE" => "SCore",
                         "options" => [ "-N NAME", "-S /bin/bash", 
                                        "-pe score SGE_SLOTS",
					"-V",
                                        qq(-masterq synapse.q), 
                                        "-r y", "-j y", "-cwd" ]
					}
		 },
                 "workstations" => {
                     "id"    => "workstations",
                     "GMXRC"  => 
                         "source /sansom/fedpacks/opt/gromacs/3.2.1/i686-pc-linux-gnu/bin/GMXRC",
                     "mdrun" => "mdrun",
                     "mdrun_mpi" => "mdrun_mpi",
                     "mpirun" => "mpirun C",
                     "start_MPI" => q(#$ -pe mpi-dual),  # hack for 2-cpu runs (no sanity checks)
                     "nproc"  => 1,
                     "qsub"   => "qsub -p -10",
                     "qsystem" => {
                         "id" => "SGE",
			 "PE" => "LAM",
                         "options" => [ "-N NAME", 
                                        "-S /bin/bash", "-r y", "-j y", 
					"-cwd","-l glibc=2.3" ],
                     },
                     "prologue" => 
                         qq(function die () {\n echo "\$1"; exit 1;\n})
                         .qq(\n)
                         .qq(function write_purge () {\n local scratchdir=\$1\n)
                         .q( echo -e '#!/bin/sh\nssh '${HOSTNAME}' rm -rvf '${scratchdir}'\n' > purge.sh)
                         .qq(\n)
                         .qq( chmod u+x purge.sh)
                         .qq(\n)
                         .q(})    
                         .qq(\n)
                         .qq(: \${SGE_TASK_ID:=x})
                         .qq(\n)
                         .qq(scratchdir=/home/scratch/\${LOGNAME}/\${JOB_NAME}.\${JOB_ID}.\${SGE_TASK_ID})
                         .qq(\n)
                         .qq(mkdir -p \${scratchdir} || die "Failed to make \${scratchdir} on \$(uname -n)")
                         .qq(\n)
                         .qq(write_purge \${scratchdir})
                         .qq(\n)
			 .qq(WORKDIR=\${PWD})
			 .qq(\n)
                         .qq(cp -v \${WORKDIR}/* \${scratchdir} || die "Error copying input files.")
                         .qq(\n)
                         .qq(cd \${scratchdir} || die "Failed to cd \${scratchdir} on \$(uname -n)")
                         .qq(\n),
                      "epilogue" =>
                         qq(\n)
                         .qq(cp -v * \${WORKDIR}),
                 },
                 "workstations_test" => {
                     "id"    => "workstations_test",
                     "GMXRC"  => 
                         "source /sansom/gfio/oliver/Gromacs/Gromacs-3.2.1-LMB/i686-pc-linux-gnu/bin/GMXRC",
                     "mdrun" => "mdrun",
                     "mdrun_mpi" => "mdrun_mpi",
                     "mpirun" => "mpirun C",
                     "start_MPI" => q(#$ -pe mpi-dual),  # hack for 2-cpu runs (no sanity checks)
                     "nproc"  => 1,
                     "qsub"   => "qsub -p -10",
                     "qsystem" => {
                         "id" => "SGE",
			 "PE" => "LAM",
                         "options" => [ "-N NAME", 
                                        "-S /bin/bash", "-r y", "-j y", 
					"-cwd","-l glibc=2.3" ],
                     },
                     "prologue" => 
                         qq(function die () {\n echo "\$1"; exit 1;\n})
                         .qq(\n)
                         .qq(function write_purge () {\n local scratchdir=\$1\n)
                         .q( echo -e '#!/bin/sh\nssh '${HOSTNAME}' rm -rvf '${scratchdir}'\n' > purge.sh)
                         .qq(\n)
                         .qq( chmod u+x purge.sh)
                         .qq(\n)
                         .q(})    
                         .qq(\n)
                         .qq(: \${SGE_TASK_ID:=x})
                         .qq(\n)
                         .qq(scratchdir=/home/scratch/\${LOGNAME}/\${JOB_NAME}.\${JOB_ID}.\${SGE_TASK_ID})
                         .qq(\n)
                         .qq(mkdir -p \${scratchdir} || die "Failed to make \${scratchdir} on \$(uname -n)")
                         .qq(\n)
                         .qq(write_purge \${scratchdir})
                         .qq(\n)
			 .qq(WORKDIR=\${PWD})
			 .qq(\n)
                         .qq(cp -v \${WORKDIR}/* \${scratchdir} || die "Error copying input files.")
                         .qq(\n)
                         .qq(cd \${scratchdir} || die "Failed to cd \${scratchdir} on \$(uname -n)")
                         .qq(\n),
                      "epilogue" =>
                         qq(\n)
                         .qq(cp -v * \${WORKDIR}),
                 },
                 "oswald" => {
                     "id"    => "oswald",
                     "mdrun" => "mdrun",
                     "mdrun_mpi" => "mdrun_mpi",
                     "mpirun" => 
                         qq(/usr/local/mpich-1.2.5..10-gcc2.96/bin/mpirun.ch_gm -machinefile gm_file.\${JOB_ID}),
                     "GMXRC"  => 
                         qq(source \${HOME}/local/gromacs/3.2.1/i686-pc-linux-gnu/bin/GMXRC),
                     "nproc"  => 2,
                     "qsub"   => "qsub",
                     "qsystem" => {
                         "id" => "SGE",
			 "PE" => "gm",
                         "options" => [ "-N NAME", "-S /bin/bash", 
                                        "-pe mpichgm NPROC",
                                        "-l h_rt=48:0:0", 
                                        qq(-m be -M oliver\@biop.ox.ac.uk),
                                        "-r y", "-j y", "-cwd" ],
                     },
                     "prologue" => 
                         qq(\n)
                         .qq(export LD_LIBRARY_PATH=\${LD_LIBRARY_PATH}:/usr/local/gm-1.6.4/lib)
                         .qq(\n),
                 },
                 "zuse" => {
                     "id"    => "zuse",
                     "mdrun" => "mdrun",
                     "mdrun_mpi" => qq(\$(which mdrun_mpi)),
                     "mpirun" => 
                         qq(/opt/scali/bin/mpirun -np \${NSLOTS} -machinefile \${TMPDIR}/machines),
                     "GMXRC"  => 
                         qq(source \${HOME}/local/gromacs/3.2.1/x86_64-unknown-linux-gnu/bin/GMXRC),
                     "nproc"  => 2,
                     "qsub"   => "qsub",
                     "qsystem" => {
                         "id" => "SGE",
			 "PE" => "scali",
                         "options" => [ "-N NAME", "-S /bin/bash", 
                                        "-pe mpi NPROC",
                                        "-l h_rt=48:0:0", 
                                        qq(-m be -M oliver\@biop.ox.ac.uk),
                                        "-r y", "-j y", "-cwd" ],
                     },
                 },
                 "neuron" => {
                     "id"    => "neuron",
                     "mdrun" => "mdrun_mpi",
                     "mdrun_mpi" => qq(\$(which mdrun_mpi)),
                     "mpirun" => 
			 qq(/opt/score/bin/scout -wait -F \${HOME}/.score/ndfile.\${JOB_ID}.\${SGE_TASK_ID} -e /tmp/scrun.\${JOB_ID}.\${SGE_TASK_ID} -nodes=NODESxPROCPERNODE, ),       
                     "GMXRC"  => 
                         qq(source /packages/opt/gromacs/3.2.1/x86_64-unknown-linux-gnu/bin/GMXRC),
                     "nproc"  => 4,
                     "procpernode" => 4,
                     "qsub"   => "qsub",
                     "qsystem" => {
                         "id" => "SGE",
			 "PE" => "SCore",
                         "options" => [ "-N NAME", "-S /bin/bash", 
                                        "-pe score SGE_SLOTS",
					"-V",
                                        qq(-masterq master.q\@neuron.bioch.ox.ac.uk), 
                                        "-r y", "-j y", "-cwd" ],
			 "prologue" => qq(\# hack for broken score_start.sh\n)
			     .qq(\# save nodefile for a task\n)
			     .qq(\# now repeat stupid   /usr/local/sge6.0/streamline/mpi/score_start.sh\n)
			     .qq(SERVER=neuron; score_dir=\${HOME}/.score;\n)
			     .qq(job_id=\${JOB_ID}.\${SGE_TASK_ID};\n)
			     .qq(ostype=\`/opt/score/deploy/config.nickname\`;\n)
			     .qq(score_bin=/opt/score/bin/bin.\${ostype};\n)
			     .qq(cat \${PE_HOSTFILE} | cut -f 1 -d " " | grep -v \${SERVER} > \${score_dir}/ndfile.\${job_id};\n)
			     .qq(echo "# Node file \${score_dir}/ndfile.\${job_id}:";\n)
			     .qq(cat \${score_dir}/ndfile.\${job_id};\n\n)
			     .qq(rm -f /tmp/scout.\${job_id};\n)
			     .qq(rm -f /tmp/scrun.\${job_id};\n)
			     .qq(ln -s \${score_bin}/scout.exe /tmp/scout.\${job_id};\n)
			     .qq(ln -s \${score_bin}/scrun.exe /tmp/scrun.\${job_id};\n\n\n),
			 "epilogue" => qq(\n\# our own cleanup\n)
			     .qq(rm -f \${score_dir}/ndfile.\${job_id};\n)
			     .qq(rm -f /tmp/scout.\${job_id};\n)
			     .qq(rm -f /tmp/scrun.\${job_id};\n),
                     },
                 },
                 "local" => {
                     "id"    => "local",
                     "mdrun" => "mdrun",
                     "mdrun_mpi" => "mdrun_mpi",
                     "mpirun" => "mpirun C -np NPROC",
                     "start_MPI" => 
                         "export PATH=/sansom/fedpacks/bin:\${PATH}; lamboot /sansom/fedpacks/etc/lam-dual.def",
                     "GMXRC" =>
                         "source /sansom/fedpacks/opt/gromacs/3.2.1/i686-pc-linux-gnu/bin/GMXRC",
                     "nproc"  => 1,
                     "qsystem" => {
                         "id" => "none",
                     },
                 },
);


my $convfac;              # holds unit conversion factors

my %Window ;
# centre of the window is taken as key
# $Window{z}->{k}    umbrella constant in kJ mol^-1nm^-2
# $Window{z}->{conf} starting configuration (of type %startconfig)


my $startconfigs;
# [i]->{filename}->"file"
#    ->{position}->[x,y,z]
#    ->{atom_id}->atom_id
#    ->{type}->"string"     type = Na|Cl|SOL|...  

my $sortedconfigs;
# $startconfigs but sorted by increasing coordinate of interest
# (only 1D)


&parseArgs();
&init_unitconversion($TEMPERATURE);
if ($show_help) {
    &print_help;
    exit(0);
} elsif ($show_version) {
    print "version: ".$program_version . "\n";
    exit(0);
} elsif ($show_targets) {
    &print_targetsystems;
    exit(0);
} elsif ($show_config) {
    &print_configformat;
    exit(0);
};
&open_log;
&checkArgs;

$startconfigs = &parseConfigList($CONFSTART);  # also checks that com is supplied
$sortedconfigs = &sortConfigs($startconfigs);  # shift to be relative to com

&computeWindows(\%Window);                     # window centres, indexed by q
&assignStartConfigs(\%Window,$sortedconfigs);  # match each centre q with a file and 
                                               # particle  Window->{config}
# print Dumper(\%Window); exit;
my (%qsuball,$sys);                            # one file per q-environment/q-system
foreach $sys (@TARGETSYSTEM) {
    $qsuball{$sys}->{"name"} = sprintf $QSUB_ALL, $sys;  # splice q-env into %s
    $qsuball{$sys}->{"fh"} = new FileHandle;             # submit all windows-script or array job
    $qsuball{$sys}->{"fh"}->open($qsuball{$sys}->{"name"},">") or 
	die "Cannot open '$qsuball{$sys}->{name}' for writing,";
    $qsuball{$sys}->{"Q"} = $qparams{$ScriptComponents{$sys}->{"qsystem"}->{"id"}};
    $qsuball{$sys}->{"task_id"} = 1;   # start indexing array jobs with 1

    logger(15,"array job \$qsuball{$sys}->{Q}->{arrayjob} = %i\n",
	  $qsuball{$sys}->{Q}->{arrayjob});
    logger(15,"queuing system id %s\n",$ScriptComponents{$sys}->{"qsystem"}->{"id"});

    &printf_arrayscript($sys,
			&run_script_header($ScriptComponents{$sys},$PREFIX));
    &printf_arrayscript($sys, 
			"#==========================================\n"
			."# submitting the jobs\n"
			."#------------------------------------------\n"
			."#\n");
    if ($qsuball{$sys}->{"Q"}->{"arrayjob"}) {
	&printf_arrayscript($sys,"# %s task array job: submit with\n#\n".
			    "#     qsub -t 1-%i %s\n",
			    $sys, scalar keys %Window, $qsuball{$sys}->{name});
    } else {
	&printf_arrayscript($sys,"# %i independent jobs for %s; "
			    ."submit by executing this script\n#\n".
			    "#     ./%s\n",
			    scalar keys %Window, $sys, $qsuball{$sys}->{name});
    }
    &printf_arrayscript($sys,"#\n#------------------------------------------\n\n\n");

    # hack to have task arrays in SCore executing (Streamline's score_start.sh is broken)
    # neuron/synapse
    if (exists $ScriptComponents{$sys}->{"qsystem"}->{"prologue"}){
	&printf_arrayscript($sys, $ScriptComponents{$sys}->{"qsystem"}->{"prologue"});
    }
    chmod 0755, $qsuball{$sys}->{"name"};
}


my ($q);                # generalised coordinate, indexing Windows
# $task_id = 1;        # index for array jobs
foreach $q (sort numeric keys %Window) {
    my (%run);
    my $win = $Window{$q};
    $run{"dir"} = &create_dir($win,$PROJECT,$PREFIX,$OUTFORMAT);
    $run{"ppa"} = &umbrella_ppa($win,$run{"dir"},$PREFIX,$OUTFORMAT);
    $run{"ndx"} = &index_ndx($win,$INDEX,$run{"dir"},$PREFIX,$OUTFORMAT);
    $run{"tpr"} = &run_tpr($win,$TOPOLOGY,$MDP,$NUM_PROCS,$run{"dir"},
                           $PREFIX,$OUTFORMAT);
    foreach $sys (@TARGETSYSTEM) {	
        &run_script($win,\%run,$ScriptComponents{$sys},$PREFIX,$OUTFORMAT);
	if ($qsuball{$sys}->{"Q"}->{"arrayjob"}) {
	    # array job: this is submitted as 'qsub -t 1-MAX'
	    # build array of directories, indexed by SGE_TASK_ID
	    &printf_arrayscript($sys,"job[%i]=%s\n",$qsuball{$sys}->{"task_id"},
				$run{"dir"}."/".$run{"script"});
	    $qsuball{$sys}->{"task_id"}++;
	} else {
	    # submit a string of jobs, just execute the file
	    &printf_arrayscript($sys,"(cd %s && %s %s)\n", 
				$run{"dir"}, $ScriptComponents{$sys}->{"qsub"},
				$run{"script"}); 
	}
    }
}

foreach $sys (@TARGETSYSTEM) {  
    if ($qsuball{$sys}->{"Q"}->{"arrayjob"}) {
	&printf_arrayscript($sys, $qsuball{$sys}->{"Q"}->{"arraycode"});

	# hack to have task arrays in SCore executing (Streamline's start-score.sh is broken)
	# neuron
	if (exists $ScriptComponents{$sys}->{"qsystem"}->{"epilogue"}){
	    &printf_arrayscript($sys, $ScriptComponents{$sys}->{"qsystem"}->{"epilogue"});
	}
    }
    $qsuball{$sys}->{"fh"}->close(); 
    msg(3,"Finished submission script for all jobs (%s): %s\n",
	$sys,$qsuball{$sys}->{"name"});
};

exit(0);
# end main


sub parseArgs () {
    &GetOptions ("help"         => \$show_help,
                 "help-targets" => \$show_targets,
                 "help-config"  => \$show_config,
                 "version"      => \$show_version,
                 "verbose"      => \$opt_verbose,
                 "debug=i"      => \$DEBUG,
                 "nwindows=i"   => \$N_WINDOWS,
                 "length=f"     => \$LENGTH,
                 "min=f"        => \$MIN,
                 "max=f"        => \$MAX,
                 "type=s"       => \$TYPE,
                 "kUmbrella=f"  => \$K_UMBRELLA,
                 "Temperature=f"=> \$TEMPERATURE,
                 "barrier=f"    => \$BARRIER, 
                 "confstart=s"  => \$CONFSTART,
                 "topology=s"   => \$TOPOLOGY,
                 "mdp=s"        => \$MDP,
                 "index=s"      => \$INDEX,
                 "np=i"         => \$NUM_PROCS,
                 "qsuball=s"    => \$QSUB_ALL,
                 "targetsystem=s@" => \@TARGETSYSTEM,
                 "Project=s"    => \$PROJECT,
                 "prefix=s"     => \$PREFIX,
                 "format=s"     => \$OUTFORMAT,
                 ) or do {
                     &print_help();
                     exit(1);
                 };
    $DEBUG = 3 if ($opt_verbose && !$DEBUG);
};

sub print_help () {
    print << "EHELP";
usage: $progname [options]

Create run input files (tpr) for umbrella sampling runs, using gromacs
(version $gmx_version). Give the number of umbrella windows and either
the length over which the windows are (equally) distributed or the minimum
and maximum coordinate. At the moment this only caters for 1D PMFs, eg just
along z in a pore or channel.

If only --length is given, the windows are create symmetrically to 0
(which is then probably the centre of mass of the reference index
group (type \`com' or \`com_t0' in the ppa)). If --min and --max or
--length and either of --min or --max are given, the window centres
are offset accordingly. The window width Dz is calculate as length/N.

Unless overriden from the commandline,the force constant is chosen so
that an ion in window i can freely diffuse into an adjacent window i+1
or i-1. The energetic barrier to this diffusion is given by
--barrier (and is the energy of the umbrella potential at z_i + 1.5*Dz). 
Set the temperature to the one set for the simulations.

This script requires all the files needed to run grompp so that it can
create the tpr files for all the windows and a special list of
starting configurations (See below). It also requires a template for
the ppa file. The grompp mdp file should have the proper run length
(eg 1 ns) and typically neither trr nor xtc output are required.

Please note (see also BUGS at the end) the following assumptions:

* Currently we assume that the direction of the sampled coordinate is
  (0,0,1), i.e. THE Z-AXIS.
* the configuration list contains absolut positions
* the center_of_mass (com) must be supplied in the conf. list
* max, min, and length are relative to the com of the ref group

Options (defaults in brackets):

--help                  this help
--help-targets          list of known target systems (for --targetsystem)
--help-config           format of the starting configuration file (for
                        --confstart)
--version               version number of $progname (quote when 
                        submitting a bug report)
--verbose               chatty output
--debug=INTEGER         set debug level (>9 switches off error checks, >11 
                        grompp is not run at all, only commandline shown)

Input files (all must exist):
--topology=FILE         gromacs topology [$TOPOLOGY]
--mdp=FILE              grompp input with run parameters  [$MDP]
--index=FILE            index file containing the reference group [$INDEX]
--confstart=FILE        list that describes available starting conformations
                        (see below for format) [$CONFSTART]
--np=INTEGER            number of processors to write tpr files for [$NUM_PROCS]

Output:
--Project=DIRECTORY     place all window directories under DIRECTORY [$PROJECT]
--prefix=STRING         output tprs are named PREFIXn.nnnn.tpr where n.nnnn 
                        is the centre of the umbrella window in nm [$PREFIX]
--format=PRINTFSTRING   pattern to create output filenames. May contain one
                        '%s' (for prefix) and one '%f' for the umbrella centre
                        (see printf(3)) [$OUTFORMAT]
--qsuball=FILE          script that contains qsub commands to submit ALL 
                        windows
--targetsystem=SYSTEM   create run script(s) for the given target systems; 
                        can be given repeatedly to produce more than one type 
                        of runscript.   [$TargetSystemDefault]
                        Use --help-targets for a list of valid SYSTEMs and 
                        the values that they define (with --verbose)

Umbrella windows:
--type=STRING           name of the particle, eg Na, Cl, NA, ... SOL (as used
                        in the input files)
--nwindows=INTEGER      number of umbrella windows [$N_WINDOWS]
--length=FLOAT          length to cover with windows, or set --min and --max
--min=FLOAT             length = abs(max - min)
--max=FLOAT 
--kUmbrella=FLOAT       force constant of the harmonic umbrella potential in
                        kJ mol^-1 nm^-2; overrides program choice
--Temperature=FLOAT     temperature of the simulation in Kelvin  [$TEMPERATURE]
--barrier=FLOAT         barrier to move to an adjacent window in kT [$BARRIER]


Bugs and Limitations
--------------------
* Only single ions as pulled groups are tested.
* conf.dat should become XML soon.
* The target system complex should become configurable through an 
  external XML file.
* The umbrella ppa file should be a template.
* Dirty hack: if there is an index file 'index.ndx' in the current directory 
  then it is automatically picked up at the grompp stage.
EHELP
};

sub print_configformat {
    print << "ECONFIGFORMAT";
=======================================================================
Format of the starting configuration file


----------------------------------------------------------------------
; comment (same as in Gromacs mdp files)
; all entries must be on one line
; '\@' lines set variables 
\@ var = value
filename   particle_type  atom_nr   pos_x pos_y pos_z 
filename   particle_type  atom_nr   pos_x pos_y pos_z
...
----------------------------------------------------------------------


filename       ::= full path to the gro or pdb file
particle_type  ::= name used in Gromacs, eg Na, Cl, SOL
atom_nr        ::= atom_id, number in gro or index file
pos_X|Y|Z      ::= coordinates of the particle

required variables: 
center_of_mass = X Y Z      ; center of mass of the reference group

optional variables:
reference_group = group     ; reference group 
reftype = com_t0|com        ; how to determine the com
RConfine = radius           ; nm
KConfine = forceconst       ; kJ/(mol*nm^2)

You can use prepconflist.pl (see 
<https://sansom.biop.ox.ac.uk/cgi-bin/cvsweb/Gromacs/Setup/UmbrellaSampling/prepconflist.pl>) for setting up the configuration list from trr frames.

==========================================================================
ECONFIGFORMAT
};

sub print_targetsystems {
    print "Valid SYSTEMs:\n";
    foreach (sort keys %ScriptComponents) {
        printf "%s ", $_;
    }
    print "\n";
    if ($DEBUG >= 3) {
        printf "\n======== Details (dumped) ========\n";
        printf "(NAME and NPROC are substitued with values.)\n";
        print Dumper(\%ScriptComponents);
        print "\n";
    }    
}


sub checkArgs {
    my $argerror = 0;
    my $opt;

    if (defined($MIN) && defined($MAX)) { 
        if (defined($LENGTH)) {
            warning("Warning: --min and --max supersede --length, ".
                  "thus --length=%f was ignored.\n", $LENGTH);
        }
        $LENGTH = abs($MAX-$MIN); 
    } elsif (defined($MIN) && ! defined($MAX)) {
        if ($LENGTH) {
            $MAX = $MIN + $LENGTH;
        } else {
            $argerror = 1;
            errmsg("You must also set one of --length or --max.\n");
        }
    } elsif (! defined($MIN) && defined($MAX)) {
        if (defined($LENGTH)) {
            $MIN = $MAX - $LENGTH;
        } else {
            $argerror = 1;
            errmsg("You must also set one of --length or --min.\n");
        }
    } elsif (defined($LENGTH)) {
        $MAX = $LENGTH/2.0;
        $MIN = - $MAX;
    } else {
        $argerror = 1;
        errmsg("You MUST set either --length or --min and --max.\n");
    };
    # now LENGTH, MIN, and MAX are set consistently
    msg(10,"checkArgs(): MIN=%f MAX=%f LENGTH=%f\n",$MIN,$MAX,$LENGTH);

    # target system(s) correct?
    if (0 == @TARGETSYSTEM) {@TARGETSYSTEM = ($TargetSystemDefault); }
    foreach (@TARGETSYSTEM) {
        if (exists $ScriptComponents{$_}) {
            $ScriptComponents{$_}->{"nproc"} = $NUM_PROCS;
            next;
	};
        $argerror = 1;
        errmsg("The SYSTEM type `%s' for --targetsystem is not defined.\n",$_);
    }

    foreach $opt (keys %requiredfiles) {
        my $file = ${$requiredfiles{$opt}};
        next if (-e $file);
        $argerror = 2;
        errmsg("Required file `%s', set with option %s,  not found.\n",
               $file,$opt);
    };

    foreach $opt (keys %requiredvars) {
        my $var = ${$requiredvars{$opt}};
        next if ($var);
        $argerror = 3;
        errmsg("Required parameter associated with option %s was not set.\n",
               $opt);
    }

    if ($argerror && $DEBUG < 10) {
        errmsg(" EXIT. Type $progname --help for the correct syntax.\n");
        exit($argerror);
    };
}

sub parseConfigList ($) {
    # better use XML::Simple
    my ($cfglist) = @_;
    my (@c,$filename,$type,$atom_id,$x,$y,$z,@comment);
    my ($key,$val);
    my $linenr = 0;
    
    open(CFG,"< $cfglist") or die "Cannot open $cfglist, stopped";
    while (<CFG>) {
        chomp;
        ++$linenr;
        next if (/^\s*;/ || /^\s*$/);      # skip comment and empty lines
        s(\s*;.*$)();                      # kill trailing whitespace and  comments
        if (/^\s*@/) {                     # We want to set a  variable...
            s(^\s*@\s*)();                 # kill @ indicator
            ($key,$val) = split /\s*=\s*/; # key = value pairs
            msg(11,"parseConfigList: key='%s' val='%s'\n",$key,$val);

            if ($configvars{$key}) {       # exists as allowed key word
                msg(10,"parseConfigList: \`$key' is a valid variable name.\n");
                $configvars{$key}->{"isset"} = 1;      # mark as read    
                if ($configvars{$key}->{"dim"} > 1) {  # array is expected
                    $UmbrellaParams{$key} = [split / /,$val];
                } else {
                    $UmbrellaParams{$key} = $val;
                }
            } else {
                warning("parseConfigList: variable \`%s' at line %i is not "
                        ."recognised (and ignored)\n",$key,$linenr);
            }
            next;
        };

        # We want to define a  configuration entry:
        ($filename,$type,$atom_id,$x,$y,$z,@comment) = split;

        # should do sanity check here but we will trust the input for now...
        push @c, { "filename" => $filename, 
                   "type" => $type, 
                   "atom_id" => $atom_id, 
                   "position" => [$x,$y,$z] };
    }
    close(CFG);


    # now check that all variables are set
    my $die_later = 0;
    foreach $key (keys %configvars) {
        if (!$configvars{$key}->{"isset"}) {
            if ($configvars{$key}->{"required"}) {
                $die_later = 1;
                errmsg("The variable \`$key' MUST be set in the configuration "
                       ."file $cfglist... this is fatal.\n");
            } else {
                # set default
                warning("parseConfigList: %s defaulted to \`%s'.\n",
                        $key,$configvars{$key}->{"default"});
                $UmbrellaParams{$key} = $configvars{$key}->{"default"};
            }
        }
    }
        
    msg(11, "parseConfigList: UmbrellaParams = %s\n",Dumper(\%UmbrellaParams));
    msg(11, "parseConfigList: startconfigs = %s\n",Dumper(\@c));
    die "Error in the configuration list $cfglist, stopped" if ($die_later);        
    return \@c;
}


sub computeWindows ($) {
    # U_i(z) = k/2 (z - z_i)**2
    #
    my ($window) = @_;     # \%window

    my $b = &convertUnits($BARRIER,"kT","kJ/mol"); # barrier in kJ/mol
    my $wfac = 35./20.;                     # b = U_i(z_i + wfac*Dz)
       # 0.035/0.020 and b = 1.5 kT used by Allen et al, BJ 84(2003),2159 
    my $K;                                  # kUmbrella in kT/Angstrom
    my $width = $LENGTH/$N_WINDOWS * 1.0;   # force float division
    my $z1 = $MIN + $width/2.0;
    my $zN = $MAX - $width/2.0;
    my ($z,$i);
    
    if (!$K_UMBRELLA) {
        $K_UMBRELLA = 2.0*$b / ($wfac*$width)**2;
    } else {
        # only for diagnostic purposes, not required 
        $BARRIER = 1.0 * $K_UMBRELLA * ($wfac*$width)**2 / 2.0;
        $b = &convertUnits($BARRIER,"kT","kJ/mol");
    };

    $K = &convertUnits(convertUnits($K_UMBRELLA,"kJ/mol","kT"),
                       "nm","Angstrom",-2);
    msg(3,"Barrier b = %g kT = %g kJ mol^-1 (T=%g K)\n",
        $BARRIER,$b,$TEMPERATURE);
    msg(3,"Window width = %.4f nm (from z1 = %g nm to z%d = %g nm)\n",
        $width, $z1, $N_WINDOWS, $zN);
    msg(3,"Using kUmbrella = %f kJ mol^-1 nm^-2 = %g kT/Angstrom^2 (T=%g K)\n",
        $K_UMBRELLA,$K,$TEMPERATURE);

    for ($i=0; $i<$N_WINDOWS; $i++) {  
        $z = $z1 + $i*$width;
        $window->{$z}->{"q"} = $z;               # generalised coordinate
        $window->{$z}->{"centre"} = [0,0,$z];    # only along z at the moment
        $window->{$z}->{"k"} = $K_UMBRELLA;      # redundant at the moment ..
        $window->{$z}->{"type"} = $TYPE;         # .. but will come in handy
    }  
}


sub sortConfigs {
    my ($c) = @_;     # $c->[i]->{filename}, ->{position}=[x,y,z], 
                      #        ->{type}=Na|water, ->{atom_id}
    my @ordered = (sort { $a->{"position"}->[$ZZ] <=> $b->{"position"}->[$ZZ]} @$c);   
    # ordered by increasing z coordinate
    # currently we only look at the z coordinates. In principle one
    # should at least calculate the vector u = max - min and then bin q
    # along u by doing the scalar product q=u.(x,y,z).

    # shift by the center_of_mass of the reference group
    foreach (@ordered) {
        my $i;
        for ($i=0;$i<3;++$i) {
            $_->{"position"}->[$i] -= $UmbrellaParams{"center_of_mass"}->[$i]
            };
    }

    if ($DEBUG > 7) {
        # get max and min for diagnostics
        my ($q_min,$q_max) = ($ordered[0]->{"position"}->[$ZZ], 
                              $ordered[-1]->{"position"}->[$ZZ]);
        my $q_range = $q_max - $q_min;
    
        msg(8,"sortConfigs: z-values ranging from %g to %g (range = %g)\n"
            . "             (relative to the Center of mass %.3f %.3f %.3f\n",
            $q_min,$q_max,$q_range, @{$UmbrellaParams{"center_of_mass"}});
    };

    return \@ordered;
}

sub numeric {
    return $a <=> $b ;
}

sub assignStartConfigs {
    my ($win,$config) = @_;
    my ($q,$closest);
    
    logger(0,"# Initial configurations\n");
    logger(0,"# z0\tz\tz-z0\tfilename\tatom_id\n");  
    foreach $q (sort numeric keys %$win) {
        msg(10,"Searching for q=%g, type=\`%s'...\n",$q,$win->{$q}->{"type"});
        $closest = &findClosestConfig($q,$win->{$q}->{"type"},$config);
        $win->{$q}->{"config"} = $closest;
        msg(10,"...found q0=%g file=%s atom_id=%i\n",
            $closest->{"position"}->[$ZZ],
            $closest->{"filename"},$closest->{"atom_id"});        
        logger(0,"%g\t%g\t%g\t%s\t%i\n",
               $q, $closest->{"position"}->[$ZZ],
               $closest->{"position"}->[$ZZ] - $q,
               $closest->{"filename"},$closest->{"atom_id"});
    }       
}

sub findClosestConfig {
    # calculate distance to each start position in turn and take
    # the one with min distance (lazy.. calculate all of them)
    # returns \% of best match 

    my ($q,$type,$config) = @_;
    my $closest;
    my @qconfig = sort 
    { abs($a->{"position"}->[$ZZ] - $q) <=> abs($b->{"position"}->[$ZZ] - $q) } @$config;
    # print "q=$q  sorted by distance:\n", Dumper(\@qconfig);
    # walk down the sorted list until we find matching type: the first hit is the best
    foreach  (@qconfig) {
	$closest = $_;    # this IS necessary (to preserve value outside loop??)
        last if ($closest->{"type"} eq $type);
    }
    if ($closest->{"type"} ne $type) {
        warning("No match for q=%f of type \`%s'; only \`%s' found.\n",
                $q,$type,$closest->{"type"});
    }
    return $closest;    
}


sub create_dir {
    my ($w,$project,$pref,$fmt) = @_;
    my $dir = sprintf("%s/".$fmt,$project,$pref,$w->{"q"});
    mkpath(["$dir"],0,0755) or warning("Failed to create dir \`$dir'.\n");
    return $dir;
}


sub pullgroup ($) {
    # canonical name for the index group
    my ($w) = @_;
    return sprintf("%s_%.3f",$w->{"type"},$w->{"q"});
}

sub umbrella_ppa {
    # the ppa file should really be a template so that the user can change it
    # Would need to define placeholders like K1_UMBRELLA or POS1X ...
    my ($w,$dir,$pref,$fmt) = @_;
    my $fn = sprintf("%s/".$fmt.".ppa",$dir,$pref,$w->{"q"});
    my $group = &pullgroup($w);
    
    open (PPA,">$fn") or die "Cannot open ppa file $fn for writing, stopped"; 
    print PPA <<"E_PPA";
; Input for umbrella sampling
; version: Gromacs 3.2.1
; see http://indigo1.biop.ox.ac.uk/wiki/index.php/Gromacs_pull_code
; created by $progname, (c) 2004 Oliver Beckstein

verbose   = yes
runtype   = umbrella       ; umbrella sampling
pulldim   = N N Y          ; restrict in Z dimension
reftype   = $UmbrellaParams{"reftype"}         ; compute center of umbrella potential
                           ; relative to initial COM of
reference_group = $UmbrellaParams{"reference_group"}    ; the reference group

; Umbrella sampling
; group which is subjected to the umbrella potential (see the index file)
group_1   = $group   
E_PPA
    printf PPA "K1        = %.2f;           ; kJ / (mol nm^2)\n", 
               $w->{"k"};
    printf PPA "Pos1      = %.3f %.3f %.3f  ; centre of the umbrella potential\n",
               @{$w->{"centre"}}; 
    if ($UmbrellaParams{"KConfine"} > 0 && $UmbrellaParams{"RConfine"} > 0.001) {
        printf PPA "; confinement potential: flat-bottomed cylinder centered on ref_com\n";
        printf PPA "KConfine = %f ; harmonic force constant in kJ/(mol*nm^2)\n",
           $UmbrellaParams{"KConfine"};
        printf PPA "RConfine = %.3f ; radius after which the potential is switched on (nm)\n",
           $UmbrellaParams{"RConfine"};
    }
    close(PPA);

    return $fn;
}


sub index_ndx {
    my ($w,$refindex,$dir,$pref,$fmt) = @_;
    my $fn = sprintf("%s/".$fmt.".ndx",$dir,$pref,$w->{"q"});

    copy ($refindex,$fn) or die "Failed to copy $refindex $fn, stopped";
    open(NDX,">> $fn") or die "Failed to append to $fn, stopped"; 
    printf NDX "[ %s ]\n", &pullgroup($w);
    printf NDX "%i\n", $w->{"config"}->{"atom_id"};
    close(NDX);
    return $fn;
}

sub run_tpr {
    my ($w,$topol,$mdp,$np,$dir,$pref,$fmt) = @_;
    my $GROMPP = $Gromacs{"grompp"};
    my $fn = sprintf("%s/".$fmt.".tpr",$dir,$pref,$w->{"q"});
    my $grompp_args = sprintf "-np %i -f %s -c %s -p %s -o %s",
        $np,$mdp, $w->{"config"}->{"filename"}, $topol, $fn;
    # quick hack to use index.ndx on the fly
    my $ndx = "index.ndx"; 
    $grompp_args.=" -n $ndx" if (-e $ndx);
    logger(8,"grompp commandline: %s\n",$grompp_args);
    $GROMPP .= " ".$grompp_args;
    if ($DEBUG < 12) {
        system($GROMPP) == 0 or die "$GROMPP failed: errorcode $?, stopped";
        unlink "mdout.mdp";   # always produced 
    } else {
        msg(12,"Debugging > 11, no grompp run. Commandline: %s\n",$GROMPP);
    }
    return $fn; 
}

# qsuball${sys}->{fh}     file handle
# qsuball${sys}->{name}   name of array script
sub printf_arrayscript  {
    my ($sys,$fmt,@args) = (@_);
    printf { $qsuball{$sys}->{fh} } $fmt, @args;
}

sub q_deffnm ($$$) {
    my ($fmt,$prefix,$q) = @_;
    return sprintf($fmt,$prefix,$q);
}

sub run_script {
    # w: window, r: run (filenames), c: scriptComponents
    my ($w,$r,$c,$prefix,$fmt) = @_;
    my $script_name = $c->{"id"}."_".$QSCRIPT;
    my $mdrun;
    my $deffnm = &q_deffnm($fmt,$prefix,$w->{"q"});
    my $mdrun_args = sprintf "-v -stepout 1000 -np %i ".
        "-s %s -pi %s -po pullout.ppa -pn %s -pd -deffnm %s",
        $c->{"nproc"}, basename($r->{"tpr"}), 
        basename($r->{"ppa"}), basename($r->{"ndx"}), $deffnm;

    $r->{"script"} = $script_name;                    # used on return by qallsub
    $r->{"runscript"} = $r->{"dir"}."/".$script_name;  
                                 
    my $qsys = $c->{"qsystem"}->{"id"};
    if ($qsys && ! exists $qparams{$qsys}) {
        errmsg("Queuing system `%s' is not known. No run script created.\n",  $qsys);
        logger(3,"Currently known queuing systems:\n%s\n",Dumper(keys %qparams));
        return;
    }

    # open run script
    $r->{"fh"} = new FileHandle;
    $r->{"fh"}->open($r->{"runscript"},">") or die "Failed to write ".$r->{"runscript"}.",stopped";
    # header (all the queuing system flags)
    print { $r->{"fh"} } &run_script_header($c,$deffnm);

    # assemble script from existing bits (xecho only prints if the arg exists)
    xecho($r->{"fh"},$c->{"prologue"});
    if  ($c->{"nproc"} > 1) {
	# TODO: Score SLOT instead of NPROC (introduce SLOT for all...)
        ($mdrun = $c->{"mpirun"}." ".$c->{"mdrun_mpi"}) =~ s(NPROC)($c->{"nproc"});
        xecho($r->{"fh"},$c->{"start_MPI"});
    } else {
        # single CPU
        $mdrun = $c->{"mdrun"};
    }    
    xecho($r->{"fh"},$c->{"GMXRC"});
    printf { $r->{"fh"} } "%s %s\n",$mdrun,$mdrun_args;
    printf { $r->{"fh"} } "gzip %s.pdo\n", $deffnm;    # save space    
    xecho($r->{"fh"},$c->{"epilogue"});

    $r->{"fh"}->close();    
    chmod 0755, $r->{"runscript"};
    msg(5,"Created runscript %s.\n",$r->{"runscript"});

}

sub run_script_header {
    # return the queuing system specific header of a run script as string
    # This function can be called from outside run_script().
    #  c: scriptComponents   qscript_name: job name displayed in SGE
    my ($c,$qscript_name) = @_;
    my $header = "";

    my $qsys = $c->{"qsystem"}->{"id"};

    # redundant when called from run_script()
    if ($qsys && ! exists $qparams{$qsys}) {
        errmsg("No header for queuing system `%s' is known.\n",  $qsys);
        logger(3,"Currently known queuing systems:\n%s\n",Dumper(keys %qparams));
        return;
    }


    # MPI (multiprocessor setup)
    # depending on parallel environment (PE) setup 
    #   NPROC:  number of processors (what you ran grompp for)
    #   NODES:  number of nodes (with PROCPERNODE  cpus per node)
    #   PROCPERNODE: cpus per node == $ScriptComponents{NAME}->{procpernode}
    #   SGE_SLOTS:  SGE slots (typically, NNODES + 1)
    #
    # this must be applied to both the header, mpirun, and mdrun_mpi
    # - mdrun_mpi is satisfied with  $c->{"nproc"}
    # - header is done below
    # - DIRTY: modify mpirun entry in the configuration...
    my ($slots,$nodes,$ncpus,$smp,$ndiv, $nrem);
    $nodes = $slots = 1;                   # default (changed if necessary, otherwise meaningless)
    $ncpus = $c->{nproc};                  # template: NPROC
    $smp = $c->{procpernode};              # template: PROCPERNODE
    if (exists $c->{qsystem}->{PE} && $c->{qsystem}->{PE} eq "SCore") {
	# distribute required cpus over nodes
	# from neuron:/usr/bin/mpirun 
	# (c) 2001-2003 Nick Birkett nrcb@streamline-computing.com
	$ndiv = int($ncpus/$smp);
	$nrem = $ncpus % $smp;
	$nodes = $ndiv + ($nrem > 0 ? 1 : 0);  # template: NODES
	$slots = $nodes + 1;                   # template: SGE_SLOTS
    } elsif (exists $c->{qsystem}->{PE} && $c->{qsystem}->{PE} eq "scali") {
	# queuing system handles distribution over nodes automatically 
	# (eg mpi on zuse)
	$slots = $c->{nproc};
    } else {
	# simple approach
	$slots = $c->{nproc};
    }
    
    # modify config entry itself (so that it can LATER be used directly)
    # (should set this up as an 'init' procedure...)
    $c->{mpirun} =~ s(NPROC)($ncpus);  
    $c->{mpirun} =~ s(PROCPERNODE)($smp);
    $c->{mpirun} =~ s(SGE_SLOTS)($slots);
    $c->{mpirun} =~ s(NODES)($nodes);


    # first line
    $header = "#!/bin/bash\n";    

    # options for queuing systems (modify and print each)
    $qscript_name = $qparams{$qsys}->{"namemangle"}($qscript_name);
    foreach (@{$c->{"qsystem"}->{"options"}}) {
        my $x;
        $x = $_;                        # dont want to change the reference
        # magical templates that are always substituted:
        $x =~ s(NAME)($qscript_name);   # modify name   
        $x =~ s(NPROC)($ncpus);  
	$x =~ s(PROCPERNODE)($smp);
	$x =~ s(SGE_SLOTS)($slots);
	$x =~ s(NODES)($nodes);
        # ugly hack for NPROC == 1 (do NOT set the proc limit; otherwise we wont 
        # land on 2cpu ws) --- what does this do? Commented out.
        #if ($x =~ /num_proc/ && $c->{"nproc"} == 1) { next };

        $header .= sprintf "%s %s\n",$qparams{$qsys}->{"prefix"}, $x;        
    }

    # say who did this...
    $header .= "# run script generated by $progname on ".date()."\n";
    $header .= sprintf "# for the '%s' environment (queuing system: %s)\n",
           $c->{"id"}, $qsys ? $qsys : "none";
    $header .= "#\n#\n";

    return $header;
}


sub xecho {
    my ($fh,$s) = @_;
    if ($s) {
        print $fh $s."\n";
    }
}

sub xechof {
    my ($fh,$fmt,@args) = @_;
    if ($fmt) {
        printf $fh $fmt."\n",@args;
    }
}


sub convertUnits ($$$$) {
    my ($x,$from,$to,$power) = @_;
    $power = 1 unless $power;
    return 1.0 * $x * ($convfac->{$from}->{$to})**($power);
}


sub init_unitconversion($) {
    my ($T) = @_;
    my $MolarGasConstant = 8.31447e-3;  # kJ mol^-1
    $convfac->{"kT"}->{"kJ/mol"} = $T * $MolarGasConstant;
    $convfac->{"kJ/mol"}->{"kT"} = 1.0/$convfac->{"kT"}->{"kJ/mol"};
    $convfac->{"nm"}->{"Angstrom"} = 10.0;
    $convfac->{"Angstrom"}->{"nm"} = 1.0/$convfac->{"nm"}->{"Angstrom"};
}

